import io
import json
import requests
import ironmill_oxide.settings as settings
from ironmill_oxide.error import OxideError

# Send the given metadata + file to the bundler to handle
def bundle(metadata: dict, circom_file: io.TextIOWrapper):
    try:
        response = requests.put(
            f'{settings.BUNDLER_URL}/add_circuit',
            files={'file': circom_file},
            params=metadata,
            # TODO auth
            )
        
        # Handle response code
        if response.status_code == 201: # Succesfuly created
            print(f'Circuit {metadata["repo_name"]} published to the registry')
        elif response.status_code == 200: # Already existed
            raise OxideError(f'A circuit named {metadata["repo_name"]} already exists on the registry')
        else:
            # TODO Error handle
            raise OxideError(f'Unexpected http response: {response}')
    except Exception as e:
        raise OxideError(e)




from ironmill_oxide.error import OxideError
import ironmill_oxide.settings as settings
import ironmill_oxide.registry as registry
import ironmill_oxide.util as util

def main(name: str):
    util.error_if_no_manifest()
    
    # Check that circuit with this name exists in registry
    if not registry.exists(name):
        raise OxideError(f'Circuit {name} does not exist in the registry')

    with open(settings.MANIFEST, 'r') as f:
        lines = f.readlines()

    # Ensure dependency section exists (Create it if no line is [dependency])
    if not (any(line.startswith(settings.DEPENDENCIES) for line in lines)):
        lines.append(f'{settings.DEPENDENCIES}\n')
    
    # Check if the given circuit exists already
    circuit_present = name in util.get_dependencies(lines)
    
    if circuit_present:
        raise OxideError(f'Circuit {name} already exists in {settings.DEPENDENCIES}')
    # Add the circuit
    else:
        for i in range(len(lines)):
            if (lines[i].startswith(settings.DEPENDENCIES)):
                print(f'Added circuit {name}')
                
                version = registry.get_version(name)
                lines.insert(i + 1, name + ' = { version = "' + version + '" }\n')
                
                break
    
    # Overwrite the file with new lines
    with open(settings.MANIFEST, 'w') as f:
        f.writelines(lines)



import re
import prompt_toolkit
from prompt_toolkit.validation import Validator
import ironmill_oxide.bundler as bundler

MAX_DESCRIPTION_CHARS = 500
MAX_NON_DESCRIPTION_CHARS = 100

def main(file):
    print('\nThis utility will walk you through specifying the details of the circuit you are publishing.')

    # Get user input for the circuit metadata
    name = prompt_toolkit.prompt(
        "package name: ",
        validate_while_typing=True,
        validator=Validator.from_callable(
            __is_valid_name,
            error_message="Not a valid package name",
            move_cursor_to_end=True,
        ),
        )

    description = prompt_toolkit.prompt("description: ",
        validate_while_typing=True,
        validator=Validator.from_callable(
            __is_valid_description,
            error_message=f"Too many characters ({MAX_DESCRIPTION_CHARS})",
            move_cursor_to_end=True,
        ),
        )

    author = prompt_toolkit.prompt("author: ",
        validate_while_typing=True,
        validator=Validator.from_callable(
            __is_valid_name,
            error_message="Not a valid author name",
            move_cursor_to_end=True,
        ),
        )

    version = prompt_toolkit.prompt(
        "version: (1.0.0) ",
        validate_while_typing=True,
        validator=Validator.from_callable(
            __is_valid_version,
            error_message="Not a valid package version",
            move_cursor_to_end=True,
        ),
        default='1.0.0',
        )

    # Bundle metadata plus circom file
    metadata = {
        'repo_name': name,
        'description': description,
        'author': author,
        'version': version,
    }

    bundler.bundle(metadata, file)

def __is_valid_name(s: str) -> bool:
    pattern = re.compile(r"^([a-z]|[0-9]|-|_)+$")
    return pattern.match(s) is not None and len(s) <= MAX_NON_DESCRIPTION_CHARS

def __is_valid_description(s: str) -> bool:
    return len(s) <= MAX_DESCRIPTION_CHARS

def __is_valid_version(s: str) -> bool:
    pattern = re.compile(r"^([0-9]+.[0-9]+.[0-9]+)+$")
    return pattern.match(s) is not None and len(s) <= MAX_NON_DESCRIPTION_CHARS



import argparse
import importlib.metadata

from ironmill_oxide.util import is_valid_circom

# NOTE Name must match this package's name
__version__ = importlib.metadata.version('ironmill-oxide')

# Create the parser
parser = argparse.ArgumentParser(description='Manage circuits')
parser.add_argument('-v', '--version', action='version', version=__version__)

# Add and document each subcommand
subparsers = parser.add_subparsers(help='sub-command help', dest='command')

""" Add """
parser_add = subparsers.add_parser('add', help='Add the <name> circuit to your manifest')
parser_add.add_argument('name', type=str, help='The name of the circuit')

""" Remove """
parser_remove = subparsers.add_parser('rm', help='Remove the <name> circuit from your manifest')
parser_remove.add_argument('name', type=str, help='The name of the circuit')

""" Init """
parser_init = subparsers.add_parser('init', help='Create a new circuits manifest')

""" Install """
parser_install = subparsers.add_parser('install', help='Install all circuits listed in your manifest')
parser_install.add_argument('-nv', '--no_verify', action='store_true', help="Skip verification of any local or new bundle hashes")

""" Search """
parser_search = subparsers.add_parser('search', help='Search the registry for any circuits that match <search_term>')
parser_search.add_argument('search_term', type=str, help='The term to search')

""" Publish """
parser_publish = subparsers.add_parser('publish', help='Publish a circuit to have it bundled and available with fields from user input')
parser_publish.add_argument('filepath', type=lambda x: is_valid_circom(parser, x), help='Path to the .circom circuit file')


# TODO compile, update, audit, doc
""" Update """
# parser_update = subparsers.add_parser('update', help='Update the lock-file to any changes to the manifest')


def main():
    return parser



import hashlib
import os.path
from ironmill_oxide import registry

# The size for each block read from the file
BLOCK_SIZE = 65536

def verify(root_dir, name):
    # Create the hash object
    hash = hashlib.sha256()
    
    # Look at each of the paths expected in the circuit's dir
    for subpath in [
        'index.js',
        f'{name}.zkey',
        f'{name}_js/{name}.wasm',
        f'{name}_js/Prover.js',
        f'{name}_js/Verifier.js',
    ]:
        path = f'{root_dir}/{subpath}'

        __update_hash(hash, path)

    return hash.hexdigest() == registry.get_hash(name)

# Update the given hash to include a hash of the contents specified by path
def __update_hash(hash, path):
    if os.path.exists(path):
        with open(path, 'rb') as f: # Open the file to read its bytes
            fb = f.read(BLOCK_SIZE) # Read from the file. Take in the amount declared above
            while len(fb) > 0: # While there is still data being read from the file
                hash.update(fb) # Update the hash
                fb = f.read(BLOCK_SIZE) # Read the next block from the file



import json
import os
from typing import List
import datetime
import semver
from zipfile import ZipFile
import io
import requests
from ironmill_oxide import hash_verifier
from ironmill_oxide.error import OxideError
import ironmill_oxide.settings as settings
import shutil

# Size in characters of the different columns
NAME_WIDTH = 20
AUTHOR_WIDTH = 10
DATE_WIDTH = 10
VERSION_WIDTH = 8

class Row:
    def __init__(self,
                 name: str,
                 description: str,
                 author: str,
                 date: str,
                 version: str,
                 ) -> None:
        self.name = name
        self.description = description
        self.author = author
        self.date = date
        self.version = version

    def __str__(self) -> str:
        result = ''
        for field, width in [
            (self.name, NAME_WIDTH),
            (self.description, Row.get_description_width()),
            (self.author, AUTHOR_WIDTH),
            # TODO Fix the formatting on date
            (self.date, DATE_WIDTH),
            (self.version, VERSION_WIDTH)
            ]:

            alignment = '{0: <' + str(width) + '}'
            
            field_s = str(field)
            if len(field_s) > width:
                field_s = field_s[0:width - 3] + '...'
            
            result += alignment.format(field_s)
            result += ' | '

        return result.rstrip(' | ')

    @staticmethod
    def get_header() -> str:
        result = ''

        for field, width in [
            ('name', NAME_WIDTH),
            ('description', Row.get_description_width()),
            ('author', AUTHOR_WIDTH),
            ('date', DATE_WIDTH),
            ('version', VERSION_WIDTH)
            ]:

            alignment = '{0: <' + str(width) + '}'

            result += alignment.format(field).upper()
            result += ' | '
        
        return result.rstrip(' | ')

    @staticmethod
    def get_description_width() -> int:
        return os.get_terminal_size().columns - NAME_WIDTH - AUTHOR_WIDTH - DATE_WIDTH - VERSION_WIDTH - 3 * 4

def search(search_term: str) -> List[Row]:
    # Send a request to the registry to search the given name
    params = {'search_term': search_term}
    url = f'{settings.REGISTRY_URL}/search'
    response = requests.get(url, params=params)
    try:
        content = json.loads(response.content)
    except json.JSONDecodeError:
        raise OxideError("Search response was invalid json")

    return list(map(lambda row: 
        Row(
            name=row.get('repo_name'),
            description=row.get('description'),
            author=row.get('author'),
            date=row.get('updated'),
            version=row.get('version'),
        )
    , content.get('results')))

# Check whether the circuit with given name exists in the registry
def exists(name: str) -> bool:
    return __get_data(name) is not None

# Get the hash for the circuit with given name
def get_hash(name: str) -> str:
    # TODO Rename to bundle_hash
    return __get_data(name).get('circuit_hash')

# Get the version for the circuit with given name
def get_version(name: str) -> str:
    return __get_data(name).get('version')

# Download and unzip the given bundle to the circuits dir
def get_bundle(name: str, verify: bool):
    params = {'repo_name': name}
    url = f'{settings.REGISTRY_URL}/get_bundle'
    response = requests.get(url, params=params)

    # Handle response code
    if not response.ok:
        raise OxideError(f'Circuit {name} does not exist in the registry')

    # Extract from the byte-stream (Without writing to a temporary zip file)
    bundle_path = f'{settings.DEP_DIR}/{name}'
    with ZipFile(io.BytesIO(response.content)) as zFile:
        zFile.extractall(path=bundle_path)

    # Verify that the given hash matches the directory's hash
    if verify and not hash_verifier.verify(bundle_path, name):
        shutil.rmtree(bundle_path)
        print(f"Deleted {bundle_path}")
        raise OxideError(f"Downloaded circuit {name}'s hash is different from the registry's hash")

# Get data for circuit with given name
def __get_data(name: str):
    params = {'repo_name': name}
    url = f'{settings.REGISTRY_URL}/get_data'
    response = requests.get(url, params=params)

    try:
        content = json.loads(response.content)
    except json.JSONDecodeError:
        raise OxideError("Search response was invalid json")
    
    return content.get('results')



import os.path
from ironmill_oxide.error import OxideError
import ironmill_oxide.settings as settings

def main():
    if os.path.exists(settings.MANIFEST):
        raise OxideError(f'{settings.MANIFEST} already exists')
    
    f = open(settings.MANIFEST, 'x')
    print(f'Create new {settings.MANIFEST} manifest')



import argparse
import os.path
from typing import List
import re
import ironmill_oxide.settings as settings
from ironmill_oxide.error import OxideError 

# Check that the given path is a valid circom file, raise a parser error if it isn't
def is_valid_circom(parser: argparse.ArgumentParser, path: str):
    if not os.path.exists(path):
        parser.error(f'The file {path} does not exist')
    elif not path.endswith('.circom'):
        parser.error(f"The file {path} isn't a circom file")
    else:
        return open(path, 'rb')


# Error if the manifest doesn't exist
def error_if_no_manifest():
    if not os.path.exists(settings.MANIFEST):
        raise OxideError(f'Could not find {settings.MANIFEST} in {os.getcwd()} or any parent directory.\nRun oxide init to generate a manifest file.')

# Return a list of all dependencies specified by the given lines
def get_dependencies(lines: List[str]) -> List[str]:
    dependencies = []
    in_dependency_section = False
    for line in lines:
        if line.startswith(settings.DEPENDENCIES):
            in_dependency_section = True
        elif line.startswith('['):
            in_dependency_section = False
        elif in_dependency_section:
            # TODO Use version, etc and not just circuit name
            pattern = r'(\w|_|-)+'
            dependencies.append(next(re.finditer(pattern, line)).group())
    
    return dependencies



import os
import shutil

from ironmill_oxide import hash_verifier
import ironmill_oxide.settings as settings
from ironmill_oxide.error import OxideError
import ironmill_oxide.util as util
import ironmill_oxide.registry as registry

def main(verify: bool):
    util.error_if_no_manifest()

    # Open the manifest to read
    with open(settings.MANIFEST, 'r') as f:
        lines = f.readlines()
    
    # Ensure circuit folder exists
    if not os.path.isdir(settings.DEP_DIR):
        os.mkdir(settings.DEP_DIR)
    
    # Get all dependency
    dependencies = util.get_dependencies(lines)

    # First ensure that each dependency is valid
    for dep in dependencies:
        if not registry.exists(dep):
            raise OxideError(f'Circuit {dep} listed in the {settings.MANIFEST} does not exist in the registry')

    # Verify existing local circuits, or download and verify any that don't exist
    amt_downloaded = 0
    for dep in dependencies:
        amt_downloaded += __verify_or_download(dep, verify)

    # Print if succesful
    print(f'Succesfully installed {amt_downloaded} packages')

def __verify_or_download(name: str, verify: bool) -> bool:
    if __is_installed(name):
        if not verify:
            return False
        else:
            # Verify that it hashes correctly
            bundle_path = f'{settings.DEP_DIR}/{name}'
            if hash_verifier.verify(bundle_path, name):
                # Don't bother to download this dependency
                print(f'Dependency is already installed: {name}')
                return False
        
        # If not, delete it and download again or error
        shutil.rmtree(bundle_path)
        print(f"Local bundle for dependency {name} doesn't match the hash listed on the registry")
        print(f"Deleted bundle {name}")
        print("Attempting to redownload...")
    
    # Download bundle from the registry
    registry.get_bundle(name, verify)
    print(f'Downloaded {name}')
    return True

# Check whether the given circuit is installed locally
def __is_installed(name: str) -> bool:
    return os.path.isdir(f'{settings.DEP_DIR}/{name}')



import os.path
import re
from ironmill_oxide.error import OxideError
import ironmill_oxide.settings as settings
import ironmill_oxide.util as util

def main(name: str):
    util.error_if_no_manifest()
    
    with open(settings.MANIFEST, 'r') as f:
        lines = f.readlines()

    # Seek the dep section, look for circuit, remove if present
    if not __attempt_rm(lines, name):
        raise OxideError(f'Circuit {name} not listed in the dependencies')
    
    # Overwrite the file with new lines
    with open(settings.MANIFEST, 'w') as f:
        f.writelines(lines)

# Attempt to remove circuit specified by name from the lines
def __attempt_rm(lines, name: str):
    for i in range(len(lines)):
        if lines[i].startswith(settings.DEPENDENCIES):
            # Go from start of dependency section to start of next section or end of file
            for j in range(i + 1, len(lines)):
                
                r = f"^{name}" + r"\b"
                if re.search(r, lines[j]) is not None:
                    print(f'Removed circuit {name}')
                    
                    lines.pop(j)

                    return True
                
                # Stop looking once dependency section ends
                elif lines[j].startswith('['):
                    break
            break
    return False


import ironmill_oxide.registry as registry

def main(search_term: str):
    # Get all search results as rows
    results = registry.search(search_term)

    # Format and print result with a header
    print(registry.Row.get_header())
    for result in results:
        print(result)



MANIFEST = 'circuits.toml'
DEPENDENCIES = '[dependencies]'
DEP_DIR = 'circuits'

_POSTPEND = '/v1/repo'
BUNDLER_URL = 'https://bundler.oxide.is' + _POSTPEND
REGISTRY_URL = 'https://registry.oxide.is' + _POSTPEND



import requests
from colorama import Fore, Style
from ironmill_oxide.error import OxideError
import ironmill_oxide.cli.get_parser as get_parser
import ironmill_oxide.cli.commands as commands

def main():
    parser = get_parser.main()
    args = parser.parse_args()
    try:
        if args.command == 'init':
            commands.init.main()

        elif args.command == 'add':
            commands.add.main(args.name)

        elif args.command == 'rm':
            commands.rm.main(args.name)

        elif args.command == 'search':
            commands.search.main(args.search_term)

        elif args.command == 'install':
            commands.install.main(not args.no_verify)

        elif args.command == 'publish':
            commands.publish.main(args.filepath)
        
    except OxideError as e:
        parser.error(f'{Fore.RED}{e}{Style.RESET_ALL}')

    except requests.exceptions.ConnectionError as e:
        parser.error(f'{Fore.RED}Connection error:\n{e}{Style.RESET_ALL}')
